
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "gbfs.h"
#include "lib.h"

#include "screen.h"
#include "basic.h"

//---------------------------------------------------------------------------
ROM_DATA ST_CMD Cmd[] = {
	{ TOKEN_PRINT,  "PRINT" },
	{ TOKEN_INPUT,  "INPUT" },
	{ TOKEN_IF,     "IF"    },
	{ TOKEN_THEN,   "THEN"  },
	{ TOKEN_GOTO,   "GOTO"  },
	{ TOKEN_FOR,    "FOR"   },
	{ TOKEN_NEXT,   "NEXT"  },
	{ TOKEN_TO,     "TO"    },
	{ TOKEN_GOSUB,  "GOSUB" },
	{ TOKEN_RETURN, "RETURN"},
	{ TOKEN_END,    "END"   },
	{ 0,            ""      },
};

//---------------------------------------------------------------------------
ST_BASIC Basic;

//---------------------------------------------------------------------------
EWRAM_CODE void BasicInit()
{
	_Memset((u8*)&Basic, 0x00, sizeof(ST_BASIC));
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicLoad(u8* pTxt, u32 size)
{
	Basic.pTxt = pTxt;
	Basic.pCur = pTxt;
	Basic.size = size;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicExec()
{
	ST_LEXER lex;
	u8* pCur;
	u16 i;

	//O(ǂ݂ăxo^܂)
	if( BasicSetLabel() == FALSE )
	{
		goto Error;
	}

	while( BasicIsCurEnd() == FALSE )
	{
		pCur = BasicGetCurPointer();
		LexerSetTokenList(&lex, pCur);

		i = 0;
		if(lex.token[i].id == TOKEN_NUMBER)
		{
			i++;
		}

		switch(lex.token[i].id)
		{
		case TOKEN_VARIABLE:
			ParseProcVariable(&lex.token[i]);
			break;

		case TOKEN_GOTO:
			//TODO
			break;

		case TOKEN_IF:
			//TODO
			break;

		case TOKEN_FOR:
			ParseProcFor(&lex.token[i]);
			break;

		case TOKEN_NEXT:
			ParseProcNext(&lex.token[i]);
			break;

		case TOKEN_PRINT:
			ParseProcPrint(&lex.token[i]);
			break;

		case TOKEN_INPUT:
			//TODO
			break;

		case TOKEN_GOSUB:
			ParseProcGosub(&lex.token[i]);
			continue;

		case TOKEN_RETURN:
			ParseProcReturn(&lex.token[i]);
			break;

		case TOKEN_END:
			goto End;
			break;

		case TOKEN_DELIMITER:
			//EMPTY
			break;

		default:
			goto Error;
			break;
		}

		BasicSetCurNextLine();
	}

End:
	KeyWait(KEY_A);
	return;

Error:
	TRACEOUT("Err: BasicExec\n");
	KeyWait(KEY_A);
}
//---------------------------------------------------------------------------
//ǂ݂ăxo^ƁAeLXg̏g[NɎ܂邩`FbNĂ܂
EWRAM_CODE bool BasicSetLabel()
{
	ST_LEXER lex;
	u8* pCur;

	while( BasicIsCurEnd() == FALSE )
	{
		pCur = BasicGetCurPointer();

		if( LexerSetTokenList(&lex, pCur) == FALSE )
		{
			return FALSE;
		}

		if(lex.token[0].id == TOKEN_NUMBER)
		{
			if( BasicAddLabel(pCur, lex.token[0].num) == FALSE )
			{
				return FALSE;
			}
		}

		BasicSetCurNextLine();
	}

	BasicSetCurRestore();

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicAddLabel(u8* pCur, s32 num)
{
	ST_BASIC* b = &Basic;
	u16 i;

	if(b->lblCnt >= BASIC_MAX_LABEL_CNT)
	{
		return FALSE;
	}

	for(i=0; BASIC_MAX_LABEL_CNT; i++)
	{
		if(b->pLblCur[i] == NULL)
		{
			b->lblNum[i]  = num;
			b->pLblCur[i] = pCur;

			b->lblCnt++;
			return TRUE;
		}

		//d`FbN
		if(b->lblNum[i] == num)
		{
			return FALSE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicIsLabel(s32 num)
{
	ST_BASIC* b = &Basic;
	u16 i;

	for(i=0; i<b->lblCnt; i++)
	{
		if(b->lblNum[i] == num)
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicSetLabelCur(s32 num)
{
	ST_BASIC* b = &Basic;
	u16 i;

	for(i=0; i<b->lblCnt; i++)
	{
		if(b->lblNum[i] == num)
		{
			b->pCur = b->pLblCur[i];
			return;
		}
	}
}
//---------------------------------------------------------------------------
//śACR+LF݂̂ɑΉĂ܂
EWRAM_CODE void BasicSetCurNextLine()
{
	ST_BASIC* b = &Basic;
	u8 chr;

	for(;;)
	{
		chr = *b->pCur++;

		if(chr == CR)
		{
			b->pCur++;		//LF 
			return;
		}

		if(chr == '\0')
		{
			b->pCur--;
			return;
		}
	}
}
//---------------------------------------------------------------------------
//NEXT߂܂Cursor|C^XLbv܂
EWRAM_CODE void BasicSetCurSkipNext()
{
	ST_LEXER lex;
	u8* pCur;

	for(;;)
	{
		BasicSetCurNextLine();

		if( BasicIsCurEnd() == TRUE )
		{
			return;
		}

		pCur = BasicGetCurPointer();
		LexerSetTokenList(&lex, pCur);

		if( (lex.token[0].id == TOKEN_NEXT) || (lex.token[1].id == TOKEN_NEXT) )
		{
			return;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* BasicGetCurPointer()
{
	return Basic.pCur;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicIsCurEnd()
{
	u32 readCnt = Basic.pCur - Basic.pTxt;

	return (readCnt >= Basic.size) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicSetCurRestore()
{
	Basic.pCur = Basic.pTxt;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicForPush(s32* pVar, s32 max)
{
	ST_BASIC* b = &Basic;

	b->pForCur[ b->forCnt ] = b->pCur;
	b->pForVar[ b->forCnt ] = pVar;
	b->forMax [ b->forCnt ] = max;

	b->forCnt++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicForPop()
{
	Basic.forCnt--;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32* BasicForGetVarPointer()
{
	return Basic.pForVar[ Basic.forCnt - 1 ];
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* BasicForGetCurPointer()
{
	return Basic.pForCur[ Basic.forCnt - 1 ];
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 BasicForGetMax()
{
	return Basic.forMax[ Basic.forCnt - 1 ];
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicForIsEmpty()
{
	return (Basic.forCnt == 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicForIsFull()
{
	return (Basic.forCnt >= BASIC_MAX_FOR_CNT-1) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicGosubPush()
{
	ST_BASIC* b = &Basic;

	b->pGosubCur[ b->gosubCnt ] = b->pCur;
	b->gosubCnt++;
}
//---------------------------------------------------------------------------
EWRAM_CODE void BasicGosubPop()
{
	ST_BASIC* b = &Basic;

	b->gosubCnt--;
	b->pCur = b->pGosubCur[ b->gosubCnt ];
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* BasicGosubGetCurPointer()
{
	return Basic.pGosubCur[ Basic.gosubCnt - 1 ];
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicGosubIsEmpty()
{
	return (Basic.gosubCnt == 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool BasicGosubIsFull()
{
	return (Basic.gosubCnt >= BASIC_MAX_GOSUB_CNT-1) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32* BasicVarGetPointer(u8* p)
{
	return &Basic.var[ *p - 'A' ];
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 BasicVarGetNum(u8* p)
{
	return Basic.var[ *p - 'A' ];
}





//---------------------------------------------------------------------------
EWRAM_CODE void LexerInit(ST_LEXER* pLex)
{
	_Memset((u8*)pLex, 0x00, sizeof(ST_LEXER));
}
//---------------------------------------------------------------------------
EWRAM_CODE bool LexerSetTokenList(ST_LEXER* pLex, u8* pStr)
{
	LexerInit(pLex);

	u16 cnt = 0 - 1;
	u8  id;

	u8  buf[CMD_MAX_LEN];
	u8* pTmp;
	u16 num;

	for(;;)
	{
		cnt++;
		pStr = LexerSkipSpace(pStr);
		id   = LexerGetId(*pStr);

		TRACEOUT("m: pStr = %x %d\n", pStr, id);

		switch(id)
		{
		case TOKEN_DELIMITER:
			LexerSetToken(pLex, cnt, id, pStr, 1);
			return TRUE;

		case TOKEN_SIGN:
			LexerSetToken(pLex, cnt, *pStr, pStr, 1);
			pStr++;
			break;

		case TOKEN_QUOTE:
			pTmp = ++pStr;

			while(*pStr != '"')
			{
				pStr++;

				if( (*pStr == CR) || (*pStr == '\0') )
				{
					return FALSE;
				}
			}
			LexerSetToken(pLex, cnt, TOKEN_STRING, pTmp, pStr - pTmp);
			pStr++;
			break;

		case TOKEN_NUMBER:
			pTmp = pStr;
			num  = 0;

			do {
				num = (num * 10) + (*pStr - '0');
				pStr++;

				if( (pStr - pTmp) >= NUM_MAX_LEN )
				{
					return FALSE;
				}

			} while( _IsDigit(*pStr) == TRUE );

			LexerSetToken(pLex, cnt, TOKEN_NUMBER, pTmp, num);
			break;

		case TOKEN_STRING:
			pTmp = pStr;
			num  = 0;

			do {
				buf[num] = _Toupper(*pStr);
				num++;
				pStr++;

				if( (pStr - pTmp) >= CMD_MAX_LEN )
				{
					return FALSE;
				}

			} while( _IsAlpha(*pStr) == TRUE );

			buf[num] = '\0';

			//ߌꂩϐ`FbN܂
			if( LexerIsCmdStr(buf) == TRUE )
			{
				LexerSetToken(pLex, cnt, LexerGetCmdId(buf), pTmp, 0);
			}
			else
			{
				if(num != 1)
				{
					return FALSE;
				}

				LexerSetToken(pLex, cnt, TOKEN_VARIABLE, pTmp, 1);
			}
			break;

		default:
			TRACEOUT("Err: %x\n", id);
			return FALSE;

		} // switch(type)

	} // for(;;)
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* LexerSkipSpace(u8* p)
{
	while(*p == ' ')
	{
		p++;
	}

	return p;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 LexerGetId(u8 chr)
{
	if( _Strchr((u8*)"+-*^/%=;(),><", chr) != NULL )
	{
		return TOKEN_SIGN;
	}
	else if( _IsDigit(chr) == TRUE )
	{
		return TOKEN_NUMBER;
	}
	else if( _IsAlpha(chr) == TRUE )
	{
		return TOKEN_STRING;
	}
	else if(chr == '"')
	{
		return TOKEN_QUOTE;
	}

	return TOKEN_DELIMITER;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool LexerIsCmdStr(u8* pStr)
{
	u16 i;

	for(i=0; Cmd[i].id != 0; i++)
	{
		if( _Strncmp(Cmd[i].str, pStr, CMD_MAX_LEN) == 0 )
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 LexerGetCmdId(u8* pStr)
{
	u16 i;

	for(i=0; Cmd[i].id != 0; i++)
	{
		if( _Strncmp(Cmd[i].str, pStr, CMD_MAX_LEN) == 0 )
		{
			return Cmd[i].id;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void LexerSetToken(ST_LEXER* pLex, u16 cnt, u8 id, u8* p, u16 num)
{
	ST_TOKEN *pTok = &pLex->token[cnt];

	pTok->id  = id;
	pTok->p   = p;
	pTok->num = num;
}





//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcVariable(ST_TOKEN* pTok)
{
	s32* p = BasicVarGetPointer(pTok->p);
	pTok++;

	if(pTok->id != '=')
	{
		return FALSE;
	}
	pTok++;

	*p = ParseGetExpr2(&pTok);
	//TODO G[`FbN

	if(pTok->id != TOKEN_DELIMITER)
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetExpr(ST_TOKEN* pTok)
{
	return ParseGetExpr2(&pTok);
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetExpr2(ST_TOKEN** ppTok)
{
	s32 r = ParseGetExpr3(ppTok);
	s32 h;
	u8  op;

	if(((**ppTok).id == '+') || ((**ppTok).id == '-'))
	{
		op = (**ppTok).id;
		*ppTok += 1;

		h = ParseGetExpr3(ppTok);
		r = ParseGetArith(op, r, h);
	}

	return r;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetExpr3(ST_TOKEN** ppTok)
{
	s32 r = ParseGetExpr4(ppTok);
	s32 h;
	u8  op;

	if(((**ppTok).id == '*') || ((**ppTok).id == '/') || ((**ppTok).id == '%'))
	{
		op = (**ppTok).id;
		*ppTok += 1;

		h = ParseGetExpr4(ppTok);
		r = ParseGetArith(op, r, h);
	}

	return r;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetExpr4(ST_TOKEN** ppTok)
{
	s32 r = ParseGetExpr5(ppTok);
	s32 h;
	u8  op;

	if((**ppTok).id == '^')
	{
		op = (**ppTok).id;
		*ppTok += 1;

		h = ParseGetExpr5(ppTok);
		r = ParseGetArith(op, r, h);
	}

	return r;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetExpr5(ST_TOKEN** ppTok)
{
	s32 ret;

	switch((**ppTok).id)
	{
	case TOKEN_VARIABLE:
		ret     = BasicVarGetNum((**ppTok).p);
		*ppTok += 1;
		break;

	case TOKEN_NUMBER:
		ret     = (**ppTok).num;
		*ppTok += 1;
		break;

	case '(':
		*ppTok += 1;
		ret = ParseGetExpr2(ppTok);

		if((**ppTok).id != ')')
		{
			//TODO G[
		}
		*ppTok += 1;
		break;

	default:
		ret = 0;
		//TODO G[
	}

	return ret;
}
//---------------------------------------------------------------------------
EWRAM_CODE s32 ParseGetArith(u8 op, s32 r, s32 h)
{
	s32 ret = 0;

	switch(op)
	{
	case '-':
		ret = r - h;
		break;

	case '+':
		ret = r + h;
		break;

	case '*':
		ret = r * h;
		break;

	case '/':
		ret = _Div(r, h);
		break;

	case '^':
		ret = _Pow(r, h);
		break;

	case '%':
		ret = _UMod(r, h);
		break;
	}

	return ret;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcFor(ST_TOKEN* pTok)
{
	//FOR
	pTok++;

	//var
	if(pTok->id != TOKEN_VARIABLE)
	{
		return FALSE;
	}
	s32* p = BasicVarGetPointer(pTok->p);
	pTok++;

	//=
	if(pTok->id != '=')
	{
		return FALSE;
	}
	pTok++;

	//expr
	*p = ParseGetExpr2(&pTok);
	//TODO G[`FbN

	//TO
	if(pTok->id != TOKEN_TO)
	{
		return FALSE;
	}
	pTok++;

	//loop cnt(expr)
	s32 cnt = ParseGetExpr2(&pTok);
	//TODO G[`FbN

	//if loop can execute at least once, push info on stack
	if(cnt >= *p)
	{
		if( BasicForIsFull() == TRUE )
		{
			return FALSE;
		}

		BasicForPush(p, cnt);
	}
	else
	{
		//otherwise, skip loop code altogether
		BasicSetCurSkipNext();
	}

	if(pTok->id != TOKEN_DELIMITER)
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcNext(ST_TOKEN* pTok)
{
	if( BasicForIsEmpty() == TRUE )
	{
		return FALSE;
	}

	//NEXT
	pTok++;

	s32  max  = BasicForGetMax();
	s32* pVar = BasicForGetVarPointer();
	*pVar += 1;

	if(*pVar > max)
	{
		BasicForPop();
	}
	else
	{
		Basic.pCur = BasicForGetCurPointer();
	}

	if(pTok->id != TOKEN_DELIMITER)
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcPrint(ST_TOKEN* pTok)
{
	s32 var;

	for(;;)
	{
		pTok++;		//1ڂ̕ PRINT

		switch(pTok->id)
		{
		case TOKEN_DELIMITER:
			pTok--;

			if((pTok->id != ';') && (pTok->id != ','))
			{
				ScreenImgNewLine();
			}
			return TRUE;

		case TOKEN_STRING:
			ScreenImgDrawStr(pTok->p, pTok->num);
			ScreenImgUpdate();
			break;

		case TOKEN_NUMBER:
		case TOKEN_VARIABLE:
			var = ParseGetExpr2(&pTok);
			//TODO G[`FbN
			pTok--;

			ScreenImgDrawNum(var);
			ScreenImgUpdate();
			break;

		case ';':
			ScreenImgSkipTab();
			break;

		case ',':
			//EMPTY
			break;

		default:
			return FALSE;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcGosub(ST_TOKEN* pTok)
{
	if( BasicGosubIsFull() == TRUE )
	{
		return FALSE;
	}
	BasicGosubPush();

	//GOSUB
	pTok++;

	//label jump(expr)
	s32 num = ParseGetExpr2(&pTok);
	//TODO G[`FbN

	if( BasicIsLabel(num) == FALSE )
	{
		return FALSE;
	}
	BasicSetLabelCur(num);

	if(pTok->id != TOKEN_DELIMITER)
	{
		return FALSE;
	}

	return TRUE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool ParseProcReturn(ST_TOKEN* pTok)
{
	if( BasicGosubIsEmpty() == TRUE )
	{
		return FALSE;
	}
	BasicGosubPop();

	//RETURN
	pTok++;

	if(pTok->id != TOKEN_DELIMITER)
	{
		return FALSE;
	}

	return TRUE;
}
