
#ifndef __BASIC_H__
#define __BASIC_H__
#ifdef __cplusplus
extern "C" {
#endif


//---------------------------------------------------------------------------
#define BASIC_MAX_LABEL_CNT					20
#define BASIC_MAX_VAR_CNT					26			//'Z' - 'A'
#define BASIC_MAX_FOR_CNT					10
#define BASIC_MAX_GOSUB_CNT					10

#define TOKEN_MAX_CNT						20

#define CMD_MAX_LEN							8
#define NUM_MAX_LEN							5

//---------------------------------------------------------------------------
enum {
	TOKEN_DELIMITER = 0x01,
	TOKEN_SIGN,
	TOKEN_VARIABLE,
	TOKEN_NUMBER,
	TOKEN_STRING,
	TOKEN_QUOTE,

	TOKEN_PRINT,
	TOKEN_INPUT,
	TOKEN_IF,
	TOKEN_THEN,
	TOKEN_GOTO,
	TOKEN_FOR,
	TOKEN_NEXT,
	TOKEN_TO,
	TOKEN_GOSUB,
	TOKEN_RETURN,
	TOKEN_END,
};

//---------------------------------------------------------------------------
typedef struct {
	u8*   pTxt;
	u8*   pCur;
	u32   size;
	//variable
	s32   var[BASIC_MAX_VAR_CNT];
	//for
	u16   forCnt;
	u8*   pForCur[BASIC_MAX_FOR_CNT];
	s32*  pForVar[BASIC_MAX_FOR_CNT];
	s32   forMax[BASIC_MAX_FOR_CNT];
	//gosub
	u16   gosubCnt;
	u8*   pGosubCur[BASIC_MAX_GOSUB_CNT];
	//label
	u16  lblCnt;
	s32  lblNum[BASIC_MAX_LABEL_CNT];
	u8*  pLblCur[BASIC_MAX_LABEL_CNT];
} ST_BASIC;


typedef struct {
	u8  id;
	u8* p;
	s32 num;
} ST_TOKEN;


typedef struct {
	ST_TOKEN token[TOKEN_MAX_CNT];
} ST_LEXER;


typedef struct {
	u8 id;
	u8 str[CMD_MAX_LEN];
} ST_CMD;

//---------------------------------------------------------------------------
EWRAM_CODE void BasicInit();
EWRAM_CODE void BasicLoad(u8* pTxt, u32 size);
EWRAM_CODE void BasicExec();

EWRAM_CODE bool BasicSetLabel();
EWRAM_CODE bool BasicAddLabel(u8* pStr, s32 num);
EWRAM_CODE bool BasicIsLabel(s32 num);
EWRAM_CODE void BasicSetLabelCur(s32 num);


EWRAM_CODE void BasicSetCurRestore();
EWRAM_CODE void BasicSetCurNextLine();
EWRAM_CODE u8*  BasicGetCurPointer();
EWRAM_CODE bool BasicIsCurEnd();

EWRAM_CODE void BasicForPush(s32* pVar, s32 max);
EWRAM_CODE void BasicForPop();
EWRAM_CODE s32* BasicForGetVarPointer();
EWRAM_CODE u8*  BasicForGetCurPointer();
EWRAM_CODE s32  BasicForGetMax();
EWRAM_CODE bool BasicForIsEmpty();
EWRAM_CODE bool BasicForIsFull();

EWRAM_CODE void BasicGosubPush();
EWRAM_CODE void BasicGosubPop();
EWRAM_CODE u8*  BasicGosubGetCurPointer();
EWRAM_CODE bool BasicGosubIsEmpty();
EWRAM_CODE bool BasicGosubIsFull();


EWRAM_CODE s32* BasicVarGetPointer(u8* p);
EWRAM_CODE s32  BasicVarGetNum(u8* p);


//----
EWRAM_CODE void LexerInit(ST_LEXER* pLex);
EWRAM_CODE bool LexerSetTokenList(ST_LEXER* pLex, u8* pStr);
EWRAM_CODE void LexerSetToken(ST_LEXER* pLex, u16 cnt, u8 id, u8* p, u16 num);

EWRAM_CODE u8   LexerGetId(u8 chr);
EWRAM_CODE u8   LexerGetCmdId(u8* pStr);
EWRAM_CODE bool LexerIsCmdStr(u8* pStr);

EWRAM_CODE u8*  LexerSkipSpace(u8* p);


//----
EWRAM_CODE bool ParseProcFor(ST_TOKEN* pTok);
EWRAM_CODE bool ParseProcNext(ST_TOKEN* pTok);
EWRAM_CODE bool ParseProcPrint(ST_TOKEN* pTok);
EWRAM_CODE bool ParseProcGosub(ST_TOKEN* pTok);
EWRAM_CODE bool ParseProcReturn(ST_TOKEN* pTok);
EWRAM_CODE bool ParseProcVariable(ST_TOKEN* pTok);

EWRAM_CODE s32  ParseGetExpr(ST_TOKEN* pTok);
EWRAM_CODE s32  ParseGetExpr2(ST_TOKEN** ppTok);
EWRAM_CODE s32  ParseGetExpr3(ST_TOKEN** ppTok);
EWRAM_CODE s32  ParseGetExpr4(ST_TOKEN** ppTok);
EWRAM_CODE s32  ParseGetExpr5(ST_TOKEN** ppTok);
EWRAM_CODE s32  ParseGetArith(u8 op, s32 r, s32 h);


#ifdef __cplusplus
}
#endif
#endif
