
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "ad_arm.h"
#include "gbfs.h"
#include "lib.h"

#include "screen.h"
#include "basic.h"
#include "main.h"

//#include "header\font\BGFont.h"


//Mode3

//---------------------------------------------------------------------------
//lib.c
extern ST_FONT Font;

//---------------------------------------------------------------------------
ST_SELECT Select;

u8 State;
//---------------------------------------------------------------------------
EWRAM_CODE int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:   StateInit();   break;
			case STATE_RESET:  StateReset();  break;
			case STATE_SELECT: StateSelect(); break;
			case STATE_SCRIPT: StateScript(); break;

			default:
				DebugMessage("[Err] main State");
				break;
		}
	}


	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateInit()
{
	InitWS();
	InitLib();
	InitKey();
	InitIRQ();
	InitFont();
	InitSprite();

	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	InitGBFS();

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitWS()
{
	REG_WSCNT = 0x4317;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitLib()
{
	LibInit();
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitKey()
{
	KeyInit();
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitIRQ()
{
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER2;

	REG_TM2D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM2CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;

	REG_IME      = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitFont()
{
	ST_FONT_SRC half;		//p

	half.pDat   = (u8*)&mplus_s10rBitmap;
	half.pSheet = (u16*)&mplus_sfnt_txt;
	half.imgCx  = MPLUS_S10R_IMG_CX;
	half.cnt    = MPLUS_S10R_FONT_CNT;
	half.cx     = MPLUS_S10R_FONT_CX;
	half.cy     = MPLUS_S10R_FONT_CY;

	Mode3SetFont(NULL, &half);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitSprite()
{
	SpriteInit();

	SpriteSetPalData16((u16*)&sprPal);
	SpriteSetData((u16*)&sprTiles, sprTilesLen);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitGBFS()
{
	if( GBFSInit() == FALSE )
	{
		SystemError("[Err] GBFS Not Found.", NULL);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateReset()
{
	SelectInit();
	ScreenImgInit();
	ScreenCurInit();

	State = STATE_SELECT;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateSelect()
{
	SelectProcMenu();

	State = STATE_SCRIPT;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateScript()
{
	u8* pFile = GBFSGetFilePointer2(Select.num);
	u32 size  = GBFSGetFileSize();

	BasicInit();
	BasicLoad(pFile, size);
	BasicExec();

	State = STATE_RESET;
}





//---------------------------------------------------------------------------
EWRAM_CODE void SelectInit()
{
	ST_SELECT* p = &Select;

	p->maxCnt = GBFSGetFileCnt();
	p->num    = 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void SelectDrawFileList()
{
	ST_SELECT* p = &Select;
	ST_FONT*   f = &Font;
	u16 i;

	for(i=0; i<p->maxCnt; i++)
	{
		GBFSGetFilePointer2(i);

		Mode3DrawFontStr(16, f->half.cy * i, GBFSGetFileName(), COLOR_WHITE);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SelectProcMenu()
{
	ST_SELECT* p = &Select;
	ST_FONT*   f = &Font;
	u16 key;

	SelectDrawFileList();
	ScreenImgUpdate();
	ScreenCurMove(0, p->num * f->half.cy);

	for(;;)
	{
		WaitForVsync();
		key = KeyGet();

		if(key == KEY_NONE)
		{
			continue;
		}

		if(key & KEY_A)
		{
			break;
		}
		else if(key & KEY_UP)
		{
			if(p->num == 0) continue;
			p->num--;

			ScreenCurMove(0, p->num * f->half.cy);
		}
		else if(key & KEY_DOWN)
		{
			if(p->num+1 >= p->maxCnt) continue;
			p->num++;

			ScreenCurMove(0, p->num * f->half.cy);
		}
	}

	KeyWait3();		//͂ɂȂ܂Ń[v܂

	ScreenCurMove(0, SCREEN_CY);
	ScreenImgClear();
	ScreenImgUpdate();
}
//---------------------------------------------------------------------------
EWRAM_CODE void SystemError(char* msg1, char* msg2)
{
	LibMode3SetScreenBuffer((u16*)VideoBuffer);

	if(msg1 != NULL) Mode3DrawFontStr(0,  0, (u8*)msg1, COLOR_WHITE);
	if(msg2 != NULL) Mode3DrawFontStr(0, 11, (u8*)msg2, COLOR_WHITE);

	for(;;){}
}
